import re
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat
)
from typing import Union, Optional, Literal, Dict


class Arg(BaseModel):
    layer_type: Literal["arg"]="arg"
    name: StrictStr
    default: Optional[Union[StrictStr, StrictInt, StrictBool, StrictFloat]]=None

    def to_string(self) -> str:

        default = self.default
        if isinstance(default, str):
            default = f'"{default}"'
    
        if default:
            return f'ARG {self.name}={default}'
        
        return f'ARG {self.name}'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('ARG ', '', line, count=1)
        token = line.strip()

        options: Dict[
            str,
            str | int | float | bool
        ] = {}

        if '=' in token:
            name, default = token.split('=')

            options['name'] = name
            options['default'] = default

        else:
            options['name'] = token

        return Arg(
            **options
        )


import re
from pydantic import (
    BaseModel,
    StrictStr
)

from typing import Optional, Literal


class User(BaseModel):
    layer_type: Literal["user"]="user"
    user_id: StrictStr
    group_id: Optional[StrictStr]=None

    def to_string(self):
        if self.group_id:
            return f'USER {self.user_id}:{self.group_id}'
        
        return f'USER {self.user_id}'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        line = re.sub(
            'USER ', 
            '', 
            line,
            count=1
        ).strip()

        user_id, group_id = cls._match_permissions(line)

        return User(
            user_id=user_id,
            group_id=group_id
        )

    @classmethod
    def _match_permissions(
        cls,
        token: str
    ):

        if ':' in token:
            user_id, group_id = token.split(':')

            return (
                user_id,
                group_id
            )
        
        else:
            return (
                token,
                None
            )


import importlib
import importlib.util
import ntpath
import os
import pathlib
import re
import sys
import tarfile
from typing import Any, Callable, Dict, List, Literal, Optional, Set, Tuple, Union

from .directive import Directives
from .layers import (
    Add,
    Arg,
    Cmd,
    Copy,
    Entrypoint,
    Env,
    Expose,
    Healthcheck,
    Label,
    Maintainer,
    OnBuild,
    Run,
    Shell,
    Stage,
    StopSignal,
    User,
    Volume,
    Workdir,
)
from .layers.mount_types import (
    BindMount,
    BindMountConfig,
    CacheMount,
    CacheMountConfig,
    SecretMount,
    SecretMountConfig,
    SSHMount,
    SSHMountConfig,
    TMPFSMount,
    TMPFSMountConfig,
)
from .memory_file import MemoryFile
from .queries import LayerQuery


class Image:
    def __init__(
        self,
        name: str,
        tag: str = "latest",
        filename: Optional[str] = None,
        path: Optional[str] = None,
    ) -> None:
        self.name = name.lower()
        self.tag = tag
        self.path = path
        self.files: List[str] = []

        if filename is None:
            stub = name.replace("/", ".")
            filename = f"Dockerfile.{stub}"

        self.filename = filename
        self._layers: List[
            Add
            | Arg
            | Cmd
            | Copy
            | Entrypoint
            | Env
            | Expose
            | Healthcheck
            | Label
            | Maintainer
            | OnBuild
            | Run
            | Shell
            | Stage
            | StopSignal
            | User
            | Volume
            | Workdir
        ] = []

        self._mount_types: Dict[
            str,
            Callable[
                [
                    BindMountConfig
                    | CacheMountConfig
                    | SecretMountConfig
                    | SSHMountConfig
                    | TMPFSMountConfig
                ],
                Union[BindMount, CacheMount, SecretMount, SSHMount, TMPFSMount],
            ],
        ] = {
            "bind": lambda config: BindMount(**config),
            "cache": lambda config: CacheMount(**config),
            "secret": lambda config: SecretMount(**config),
            "ssh": lambda config: SSHMount(**config),
            "tmpfs": lambda config: TMPFSMount(**config),
        }

        self.directives = Directives()
        self._variable_template_pattern = re.compile(r"(\$\{)([^\}]+)\}|(\$)([\w]+)")

        self._variable_resolved_name_pattern = re.compile(
            r"(?<=\$\{).+?(?=:-|\})|(?<=\$)([\w]+)"
        )

        self._variable_clean_pattern = re.compile(r"\$|\{|\}")

    @property
    def full_name(self):
        return f"{self.name}:{self.tag}"

    def __iter__(self):
        for layer in self._layers:
            yield layer

    @classmethod
    def import_from_file(cls, filepath: str):
        path = pathlib.Path(filepath).absolute().resolve()
        path_str = str(path)

        package_dir = path.parent
        package_dir_path = str(package_dir)
        package_dir_module = package_dir_path.split("/")[-1]

        package = ntpath.basename(path)
        package_slug = package.split(".")[0]

        spec = importlib.util.spec_from_file_location(
            f"{package_dir_module}.{package_slug}", path_str
        )

        if path_str not in sys.path:
            sys.path.append(str(package_dir.parent))

        module = importlib.util.module_from_spec(spec)
        sys.modules[module.__name__] = module

        spec.loader.exec_module(module)

        return {
            name: value
            for name, value in module.__dict__.items()
            if isinstance(value, Image)
        }

    @classmethod
    def generate_from_file(cls, filepath: str, output_path: Optional[str] = None):
        layers: List[
            Add
            | Arg
            | Cmd
            | Copy
            | Entrypoint
            | Env
            | Expose
            | Healthcheck
            | Label
            | Maintainer
            | OnBuild
            | Run
            | Shell
            | Stage
            | StopSignal
            | User
            | Volume
            | Workdir
        ] = []

        directives = Directives()
        filename = pathlib.Path(filepath).name

        with open(filepath) as dockerfile:
            layers.extend(directives.parse(dockerfile.readlines()))

            image_sources: Stage = [
                layer for layer in layers if layer.layer_type == "stage"
            ]

            image_source: Stage = image_sources[-1]

            if output_path:
                filename = output_path

            image = Image(
                image_source.base,
                tag=image_source.tag,
                filename=filename,
                path=output_path,
            )

            image.from_layers(layers)

            return image

    @classmethod
    def generate_from_string(
        cls,
        dockerfile: str | bytes | List[str] | List[bytes],
        filename: Optional[str] = None,
        output_path: Optional[str] = None,
    ):
        layers: List[
            Add
            | Arg
            | Cmd
            | Copy
            | Entrypoint
            | Env
            | Expose
            | Healthcheck
            | Label
            | Maintainer
            | OnBuild
            | Run
            | Shell
            | Stage
            | StopSignal
            | User
            | Volume
            | Workdir
        ] = []

        directives = Directives()

        layers.extend(directives.parse(dockerfile))

        image_sources: List[Stage] = [
            layer for layer in layers if layer.layer_type == "stage"
        ]

        if len(image_sources) < 1:
            return Image("Unknown", path=output_path)

        image_source = image_sources[-1]

        if output_path:
            filename = output_path

        image = Image(
            image_source.base, tag=image_source.tag, filename=filename, path=output_path
        )

        image.from_layers(layers)

        return image

    def from_string(self, dockerfile: str):
        self._layers.extend(self.directives.parse(dockerfile))

        return self

    def from_file(self, filepath: str):
        filename = pathlib.Path(filepath).name
        self.filename = filename

        with open(filepath) as dockerfile:
            self._layers.extend(self.directives.parse(dockerfile.readlines()))

        return self

    def from_layers(
        self,
        layers: List[
            Add
            | Arg
            | Cmd
            | Copy
            | Entrypoint
            | Env
            | Expose
            | Healthcheck
            | Label
            | Maintainer
            | OnBuild
            | Run
            | Shell
            | Stage
            | StopSignal
            | User
            | Volume
            | Workdir
        ] = [],
    ):
        self._layers = layers

        return self

    def layers(
        self,
        layer_types: Optional[
            Literal[
                "add",
                "arg",
                "cmd",
                "copy",
                "entrypoint",
                "env",
                "expose",
                "healthcheck",
                "label",
                "maintainer",
                "onbuild",
                "run",
                "shell",
                "stage",
                "stopsignal",
                "user",
                "volume",
                "workdir",
            ]
            | List[
                Literal[
                    "add",
                    "arg",
                    "cmd",
                    "copy",
                    "entrypoint",
                    "env",
                    "expose",
                    "healthcheck",
                    "label",
                    "maintainer",
                    "onbuild",
                    "run",
                    "shell",
                    "stage",
                    "stopsignal",
                    "user",
                    "volume",
                    "workdir",
                ]
            ]
        ] = None,
        attribute: Optional[Tuple[str, Any]] = None,
    ):
        if layer_types is not None and not isinstance(layer_types, list):
            layer_types = [layer_types]

        if isinstance(layer_types, list):
            layers = []
            for layer_type in layer_types:
                layers.extend(
                    [layer for layer in self._layers if layer.layer_type == layer_type]
                )

            return LayerQuery(layers)

        if attribute:
            attribute_name, attribute_value = attribute
            layers = LayerQuery(self._layers)
            return layers.get(attribute_name, attribute_value)

        return LayerQuery(self._layers)

    def resolve(self, defaults: Dict[str, str] = {}, skip: List[str] = []):
        resolved_args = self.get_resolved_args()
        resolved_args.update(defaults)

        resolved_args = {
            arg: value for arg, value in resolved_args.items() if arg not in skip
        }

        layers = list(self._layers)

        for idx, layer in enumerate(layers):
            if isinstance(layer, Arg):
                layer_data = layer.model_dump()
                layer_data["default"] = resolved_args.get(layer.name, layer.default)

                layers[idx] = Arg(**layer_data)

            elif isinstance(layer, Env):
                values = []
                for key_idx, key in enumerate(layer.keys):
                    values.append(resolved_args.get(key, layer.values[key_idx]))

                layer_data = layer.model_dump()
                layer_data["values"] = values
                layers[idx] = Env(**layer_data)

        image = Image(self.name, tag=self.tag, filename=self.filename, path=self.path)

        image.from_layers(layers)

        resolved_layers: List[
            Add
            | Arg
            | Cmd
            | Copy
            | Entrypoint
            | Env
            | Expose
            | Healthcheck
            | Label
            | Maintainer
            | OnBuild
            | Run
            | Shell
            | Stage
            | StopSignal
            | User
            | Volume
            | Workdir
        ] = []

        for layer in image:
            if not isinstance(layer, (Arg, Env)):
                layer = self._resolve_layer(layer, resolved_args)

            resolved_layers.append(layer)

        image.from_layers(resolved_layers)

        return image

    def get_resolved_args(self):
        arg_values: Dict[str, Any] = {}
        arg_layers = self.layers(layer_types="arg")
        for arg in arg_layers:
            arg_name = re.sub(self._variable_clean_pattern, "", arg.name)

            if (
                arg_values.get(arg_name) is None
                and arg.default
                and (not re.search(self._variable_template_pattern, arg.default))
            ):
                arg_values[arg_name] = arg.default

        resolved_args: Dict[str, Any] = {}
        for arg_name in arg_values:
            resolved_args[arg_name] = self._reduce_args(arg_name, arg_values)

        for arg_name, value in resolved_args.items():
            for match in re.finditer(self._variable_template_pattern, value or ""):
                arg_value: str | None = None
                if arg_variable := re.search(
                    self._variable_resolved_name_pattern, match.group(0)
                ):
                    arg_value = resolved_args.get(arg_variable.group(0))

                if arg_value:
                    value = re.sub(match.group(0), arg_value, value)

                resolved_args[arg_name] = value

        env_values: Dict[str, Any] = {}
        env_layers = self.layers(layer_types="env")
        for env in env_layers:
            for key, value in zip(env.keys, env.values):
                if env_values.get(key) and isinstance(value, str):
                    for match in re.finditer(self._variable_template_pattern, value):
                        arg_value: str | None = None
                        if envar_variable := re.search(
                            self._variable_resolved_name_pattern, match.group(0)
                        ):
                            arg_value = resolved_args.get(envar_variable.group(0))

                        if arg_value:
                            value = arg_value

                if not re.search(self._variable_template_pattern, value):
                    env_values[key] = value

        resolved_args.update(env_values)

        for arg_name, value in resolved_args.items():
            resolved_args.update(self._resolve_args(arg_name, value, resolved_args))

        return resolved_args

    def _resolve_layer(
        self,
        layer: (
            Add
            | Cmd
            | Copy
            | Entrypoint
            | Expose
            | Healthcheck
            | Label
            | Maintainer
            | OnBuild
            | Run
            | Shell
            | Stage
            | StopSignal
            | User
            | Volume
            | Workdir
        ),
        resolved_args: Dict[str, Any],
    ):
        model_data = layer.model_dump()
        for field, field_value in model_data.items():
            if isinstance(field_value, str):
                matches: Set[str] = set()
                for match in re.finditer(self._variable_template_pattern, field_value):
                    if resolved_name := re.search(
                        self._variable_resolved_name_pattern, match.group(0)
                    ):
                        matches.add((resolved_name.group(0), match.group(0)))

                matches = list(matches)

                for resolved_name, match in matches:
                    if resolved_value := resolved_args.get(resolved_name):
                        field_value = re.sub(
                            r"\$\{"
                            + f"{resolved_name}?"
                            + r"\}|\$"
                            + f"{resolved_name}",
                            resolved_value,
                            field_value,
                        )

                model_data[field] = field_value

            elif isinstance(field_value, list):
                for idx, value in enumerate(field_value):
                    if isinstance(value, str):
                        matches: Set[str] = set()
                        for match in re.finditer(
                            self._variable_template_pattern, value
                        ):
                            if resolved_name := re.search(
                                self._variable_resolved_name_pattern, match.group(0)
                            ):
                                matches.add((resolved_name.group(0), match.group(0)))

                        matches = list(matches)

                        for resolved_name, match in matches:
                            if resolved_value := resolved_args.get(resolved_name):
                                value = re.sub(
                                    r"\$\{"
                                    + f"{resolved_name}?"
                                    + r"\}|\$"
                                    + f"{resolved_name}",
                                    resolved_value,
                                    value,
                                )

                        model_data[field][idx] = value

        return layer.model_copy(update=model_data)

    def _reduce_args(self, arg_name: str, args: Dict[str, Any]):
        arg_value = args.get(arg_name)

        if arg_value in args:
            return self._reduce_args(arg_value, args)

        return arg_value

    def _resolve_args(self, arg_name: str, value: str, resolved: Dict[str, str]):
        if value is None:
            return resolved

        resolved_items = [
            (resolved_name, resolved_value)
            for resolved_name, resolved_value in resolved.items()
            if resolved_value and resolved_value != value
        ]

        for resolved_arg, resolved_value in resolved_items:
            for match in re.finditer(r"(\$)([\w]+)", resolved_value):
                match_key = re.sub(r"\$|", "", match.group(0))

                if arg_name == match_key:
                    resolved[resolved_arg] = resolved_value.replace(
                        match.group(0), value
                    )

            for match in re.finditer(r"(\$\{)([\w]+)(\})", resolved_value):
                match_key = re.sub(self._variable_clean_pattern, "", match.group(0))

                if arg_name == match_key:
                    resolved[resolved_arg] = resolved_value.replace(
                        match.group(0), value
                    )

        return resolved

    def to_string(self) -> str:
        return "\n\n".join([layer.to_string().strip("\n") for layer in self._layers])

    def to_context(self) -> MemoryFile:
        if os.path.exists(self.filename) is False:
            self.to_file()

        image_file = MemoryFile(self.to_string().encode(), name=self.filename)

        image_file.seek(0)

        tar_file_number = image_file.file_number + 1

        tar_file = MemoryFile(b"", file_number=tar_file_number)

        context = tarfile.open(fileobj=tar_file, mode="w")

        image_file_info = context.gettarinfo(fileobj=image_file)

        context.addfile(image_file_info, image_file)

        for file in self.files:
            filename = pathlib.Path(file).name

            with open(file, "rb") as context_file:
                info = context.gettarinfo(fileobj=context_file, arcname=filename)

                context.addfile(info, context_file)

        context.close()
        tar_file.seek(0)

        return tar_file

    def to_file(self, filepath: Optional[str] = None):
        if filepath is None:
            filepath = self.filename

        elif filepath != self.filename:
            self.filename = filepath

        self.files.append(filepath)

        with open(filepath, "w") as dockerfile:
            dockerfile.write(self.to_string())

    def to_memory_file(self, filepath: Optional[str] = None):
        if filepath is None:
            filepath = self.filename

        elif filepath != self.filename:
            self.filename = filepath

        self.files.append(filepath)

        with open(filepath, "w") as dockerfile:
            dockerfile.write(self.to_string())

    def clear(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)
            self._layers.clear()

    def add(
        self,
        source: str,
        destination: str,
        user_id: Optional[str] = None,
        group_id: Optional[str] = None,
        permissions: Optional[int] = None,
        checksum: Optional[str] = None,
        link: bool = False,
    ):
        self._layers.append(
            Add(
                source=source,
                destination=destination,
                user_id=user_id,
                group_id=group_id,
                permissions=permissions,
                checksum=checksum,
                link=link,
            )
        )

        return self

    def arg(self, name: str, default: Union[str, int, bool, float]):
        self._layers.append(Arg(name=name, default=default))

        return self

    def cmd(self, command: List[Union[str, int, bool, float]]):
        self._layers.append(Cmd(command=command))

        return self

    def copy(
        self,
        source: str,
        destination: str,
        user_id: Optional[str] = None,
        group_id: Optional[str] = None,
        permissions: Optional[Literal[r"^[0-7]*$"]] = None,
        from_layer: Optional[str] = None,
        link: bool = False,
    ):
        copy_layer = Copy(
            source=source,
            destination=destination,
            user_id=user_id,
            group_id=group_id,
            permissions=permissions,
            from_layer=from_layer,
            link=link,
        )

        self._layers.append(copy_layer)

        self.files.append(f"./{copy_layer.source}")

        return self

    def entrypoint(self, command: List[str]):
        self._layers.append(Entrypoint(command=command))

        return self

    def env(
        self,
        keys: str | List[str],
        values: str | int | bool | float | List[str | int | bool | float],
    ):
        if isinstance(keys, (str, int, bool, float)) and isinstance(
            values, (str, int, bool, float)
        ):
            keys = [keys]
            values = [values]

        self._layers.append(Env(keys=keys, values=values))

    def expose(self, ports: List[int]):
        self._layers.append(Expose(ports=ports))

        return self

    def healthcheck(
        self,
        interval: int,
        timeout: int,
        start_period: int,
        retries: int,
        command: List[Union[str, int, bool, float]],
    ):
        self._layers.append(
            Healthcheck(
                interval=interval,
                timeout=timeout,
                start_period=start_period,
                retries=retries,
                command=Cmd(command=command),
            )
        )

        return self

    def label(self, name: str, value: str):
        self._layers.append(Label(name=name, value=value))

        return self

    def maintainer(self, author: str):
        self._layers.append(Maintainer(author=author))

        return self

    def onbuild(
        self,
        instruction: (
            Add
            | Arg
            | Cmd
            | Copy
            | Entrypoint
            | Env
            | Expose
            | Healthcheck
            | Label
            | Run
            | Shell
            | StopSignal
            | User
            | Volume
            | Workdir
        ),
    ):
        self._layers.append(OnBuild(instruction=instruction))

        return self

    def run(
        self,
        command: str,
        network: Optional[Literal["default", "host", "none"]] = None,
        security: Optional[Literal["insecure", "sandbox"]] = None,
        mount: BindMountConfig
        | CacheMountConfig
        | SecretMountConfig
        | SSHMountConfig
        | TMPFSMountConfig = None,
    ):
        mount_type: Union[
            BindMount, CacheMount, SecretMount, SSHMount, TMPFSMount, None
        ] = None

        if mount:
            mount_type_name = mount.get("mount_type", "bind")
            mount_type = self._mount_types.get(mount_type_name)(mount)

        self._layers.append(
            Run(command=command, mount=mount_type, network=network, security=security)
        )

        return self

    def shell(
        self, executable: str, parameters: Optional[List[str | int | float | bool]]
    ):
        self._layers.append(Shell(executable=executable, parameters=parameters))

        return self

    def stage(
        self,
        base: str,
        tag: str | None = None,
        alias: str | None = None,
        platform: str | None = None,
    ):
        self._layers.append(Stage(base=base, tag=tag, alias=alias, platform=platform))

        return self

    def stopsignal(self, signal: int):
        self._layers.append(StopSignal(signal=signal))

        return self

    def user(self, user_id: str, group_id: Optional[str] = None):
        self._layers.append(User(user_id=user_id, group_id=group_id))

        return self

    def volume(self, paths: List[str]):
        self._layers.append(Volume(paths=paths))

        return self

    def workdir(self, path: str):
        self._layers.append(Workdir(path=path))

        return self



import re
from pydantic import (
    BaseModel
)
from typing import Dict, Literal
from .add import Add
from .arg import Arg
from .cmd import Cmd
from .copy import Copy
from .entrypoint import Entrypoint
from .env import Env
from .expose import Expose
from .healthcheck import Healthcheck
from .label import Label
from .run import Run
from .shell import Shell
from .stopsignal import StopSignal
from .user import User
from .volume import Volume
from .workdir import Workdir


class OnBuild(BaseModel):
    layer_type: Literal['onbuild']='onbuild'
    instruction: (
        Add | 
        Arg | 
        Cmd | 
        Copy | 
        Entrypoint | 
        Env | 
        Expose | 
        Healthcheck | 
        Label |
        Run | 
        Shell |
        StopSignal |
        User |
        Volume |
        Workdir
    )

    class Config:
        arbitrary_types_allowed=True

    def to_string(self):
        instruction_command = self.instruction.to_string()
        return f'ONBUILD {instruction_command}'
    
    def parse(
        cls,
        line: str
    ):
        line = re.sub('ONBUILD ', '', line, count=1).strip()

        directives: Dict[
            Literal[
                'ADD',
                'ARG',
                'CMD',
                'COPY',
                'ENTRYPOINT',
                'ENV',
                'EXPOSE',
                'FROM',
                'HEALTHCHECK',
                'LABEL',
                'MAINTAINER',
                'ONBUILD',
                'RUN',
                'SHELL',
                'STOPSIGNAL',
                'USER',
                'VOLUME',
                'WORKDIR'
            ],
            Add |
            Arg |
            Cmd |
            Copy |
            Entrypoint |
            Env |
            Expose |
            Healthcheck |
            Label |
            OnBuild |
            Run |
            Shell |
            StopSignal |
            User |
            Volume | 
            Workdir
        ] = {
            'ADD': Add,
            'ARG': Arg,
            'CMD': Cmd,
            'COPY': Copy,
            'ENTRYPOINT': Entrypoint,
            'ENV': Env,
            'EXPOSE': Expose,
            'HEALTHCHECK': Healthcheck,
            'LABEL': Label,
            'ONBUILD': OnBuild,
            'RUN': Run,
            'SHELL': Shell,
            'STOPSIGNAL': StopSignal,
            'USER': User,
            'VOLUME': Volume,
            'WORKDIR': Workdir
        }
        
        if directive_type := re.search(
            r'ADD|ARG|CMD|COPY|ENTRYPOINT|ENV|EXPOSE|HEALTHCHECK|LABEL|RUN|SHELL|STOPSIGNAL|USER|VOLUME|WORKDIR',
            line
        ):
            directive_name: Literal[
                'ADD',
                'ARG',
                'CMD',
                'COPY',
                'ENTRYPOINT',
                'ENV',
                'EXPOSE',
                'FROM',
                'HEALTHCHECK',
                'LABEL',
                'MAINTAINER',
                'ONBUILD',
                'RUN',
                'SHELL',
                'STOPSIGNAL',
                'USER',
                'VOLUME',
                'WORKDIR'
            ] = directive_type.group(0) 
            directive = directives[directive_name]

            token = re.sub(f'${directive_name} ', '', line, count=1)

            return directive.parse(token)



        


from .image import Image
from .layers import (
    Add,
    Arg,
    Cmd,
    Copy,
    Entrypoint,
    Env,
    Expose,
    Healthcheck,
    Label,
    Maintainer,
    OnBuild,
    Run,
    Shell,
    Stage,
    StopSignal,
    User,
    Volume,
    Workdir
)


import re
from typing import Dict, Literal, Optional, Union

from pydantic import BaseModel, DirectoryPath, FilePath, StrictBool, StrictStr, constr


class Copy(BaseModel):
    layer_type: Literal["copy"] = "copy"
    source: Union[FilePath, DirectoryPath, StrictStr]
    destination: StrictStr
    user_id: Optional[StrictStr] = None
    group_id: Optional[StrictStr] = None
    permissions: Optional[constr(max_length=4, pattern=r"^[0-7]*$")] = None
    from_layer: Optional[StrictStr] = None
    link: StrictBool = False

    def to_string(self) -> str:
        copy_string = "COPY"

        if self.user_id and self.group_id:
            copy_string = f"{copy_string} --chown={self.user_id}:{self.group_id}"

        elif self.user_id:
            copy_string = f"{copy_string} --chown={self.user_id}"

        if self.permissions:
            copy_string = f"{copy_string} --chmod={self.permissions}"

        if self.from_layer:
            copy_string = f"{copy_string} --from={self.from_layer}"

        if self.link:
            copy_string = f"{copy_string} --link"

        return f"{copy_string} ./{self.source} {self.destination}"

    @classmethod
    def parse(
        cls,
        line: str,
    ):
        line = re.sub("COPY ", "", line, count=1).strip()
        tokens = line.strip("\n").split(" ")

        options: Dict[str, str | bool | int] = {}
        remainders = []

        for token in tokens:
            if re.search(r"--link", token):
                options["link"] = True

            elif re.search(r"--checksum", token):
                options["checksum"] = re.sub(r"--checksum=", "", token)

            elif re.search(r"--from", token):
                options["from_layer"] = re.sub(r"--from=", "", token)

            elif re.search(r"--chown", token):
                token = re.sub(r"--checksum=", "", token)

                (user_id, group_id, permissions) = cls._match_permissions(token)

                options["user_id"] = user_id
                options["group_id"] = group_id
                options["permissions"] = permissions

            else:
                remainders.append(token)

        if len(remainders) > 2:
            destination = remainders.pop()
            source = " ".join(remainders)

        else:
            source, destination = remainders[0], remainders[1]

        return Copy(source=source, destination=destination, **options)

    @classmethod
    def _match_permissions(cls, token: str):
        if permissions := re.search(r"[0-7]{4}|[0-7]{3}", token):
            return (None, None, permissions.group(0))

        elif ":" in token:
            user_id, group_id = token.split(":")

            return (user_id, group_id, None)

        else:
            return (token, None, None)



import re
from pydantic import (
    BaseModel,
    StrictStr,
    conlist
)
from typing import Literal


class Volume(BaseModel):
    layer_type: Literal["volume"]="volume"
    paths: conlist(StrictStr, min_length=1)

    def to_string(self):
        paths = ', '.join(self.paths)

        return f'VOLUME [{paths}]'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        line = re.sub('VOLUME ', '', line, count=1).strip()
        paths = [
            arg.strip() for arg in re.sub(
                r'\[|\]', 
                '', 
                line
            ).split(',')
        ]

        return Volume(
            paths=paths
        )


import asyncio
import inspect
import pathlib
import sys
import textwrap

from cocoa.cli import CLI

from .templates import basic


async def create_dcrx_file(
    loop: asyncio.AbstractEventLoop,
    path: str,
):
    image_file = await loop.run_in_executor(
        None,
        open,
        path,
        "w",
    )

    try:
        await loop.run_in_executor(
            None,
            image_file.write,
            textwrap.dedent(inspect.getsource(basic)),
        )

    except Exception:
        pass

    await loop.run_in_executor(None, image_file.close)


@CLI.command()
async def new(
    path: str,
    overwrite: bool = False,
):
    """
    Create a new DCRX image at the provided path

    @param path The path to create the Image file at
    @param overwrite If specified, if a file exists at the provided path it will be overwritten
    """
    loop = asyncio.get_event_loop()

    image_path = await loop.run_in_executor(
        None,
        pathlib.Path,
        path,
    )

    absolute_path = await loop.run_in_executor(None, image_path.absolute)

    resolved_path = await loop.run_in_executor(None, absolute_path.resolve)

    filepath_exists = await loop.run_in_executor(None, resolved_path.exists)

    message = f"Created new scratch Image at {path}"

    if filepath_exists and overwrite is True:
        await create_dcrx_file(
            loop,
            str(resolved_path),
        )

        await loop.run_in_executor(None, sys.stdout.write, message)

    elif filepath_exists is False:
        await create_dcrx_file(
            loop,
            str(resolved_path),
        )

        await loop.run_in_executor(None, sys.stdout.write, message)

    else:
        message = f"Err. - file at {path} already exists!"

        await loop.run_in_executor(None, sys.stdout.write, message)



import asyncio
import sys

from cocoa.cli import CLI, RawFile

from dcrx.image import Image


@CLI.command()
async def validate(
    path: RawFile[str],
):
    dockerfile = path.data
    loop = asyncio.get_event_loop()

    await loop.run_in_executor(
        None,
        sys.stdout.write,
        "Validating Dockerfile...\n",
    )

    try:
        image = Image("dcrx").from_string(dockerfile)

        await loop.run_in_executor(
            None,
            sys.stdout.write,
            "✓ Valid!\n\n",
        )

        for directive in image.to_string().split("\n"):
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                f"{directive}\n",
            )

        await loop.run_in_executor(
            None,
            sys.stdout.write,
            "\n",
        )

    except Exception as err:
        await loop.run_in_executor(
            None,
            sys.stdout.write,
            f"𐄂 Validation failed - {str(err)}\n",
        )



import re
from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool,
    constr
)

from typing import Literal, Optional, Dict, List, Any
from .arg import Arg
from .env import Env


class Add(BaseModel):
    layer_type: Literal["add"]="add"
    source: StrictStr
    destination: StrictStr
    user_id: Optional[StrictStr]=None
    group_id: Optional[StrictStr]=None
    permissions: Optional[constr(max_length=4, pattern=r'^[0-7]*$')]=None
    checksum: Optional[StrictStr]=None
    link: StrictBool=False

    def to_string(self) -> str:
        add_string = 'ADD'

        if self.user_id and self.group_id:
            add_string = f'{add_string} --chown={self.user_id}:{self.group_id}'
        
        elif self.user_id:
            add_string = f'{add_string} --chown={self.user_id}'

        if self.permissions:
            add_string = f'{add_string} --chmod={self.permissions}'
        
        if self.checksum:
            add_string = f'{add_string} --checksum={self.checksum}'

        if self.link:
            add_string = f'{add_string} --link'

        return f'{add_string} {self.source} {self.destination}'
    
    @classmethod
    def parse(
        cls,
        line: str,
    ):
        
        line = re.sub('ADD ', '', line, count=1).strip()
        tokens = line.strip('\n').split(' ')

        options: Dict[str, str | bool | int] = {}
        remainders = []

        for token in tokens:
            if re.search(
                '--link',
                token
            ):
                options['link'] = True

            elif re.search(
                '--checksum',
                token
            ):
                options['checksum'] = re.sub(
                    '--checksum=',
                    '',
                    token
                )

            elif re.search(
                '--chown',
                token
            ):
                token = re.sub(
                    '--checksum=',
                    '',
                    token
                )
                
                (
                    user_id,
                    group_id,
                    permissions
                ) = cls._match_permissions(token)

                options['user_id'] = user_id
                options['group_id'] = group_id
                options['permissions'] = permissions

            else:
                remainders.append(token)

        if len(remainders) > 2:
            destination = remainders.pop()
            source = ' '.join(remainders)

        else:
            source, destination = remainders[0], remainders[1]

        return Add(
            source=source,
            destination=destination,
            **options
        )

    @classmethod
    def _match_permissions(
        cls,
        token: str
    ):
        if permissions := re.search(
            r'[0-7]{4}|[0-7]{3}',
            token
        ):
            return (
                None,
                None,
                permissions.group(0)
            )
        
        elif ':' in token:
            user_id, group_id = token.split(':')

            return (
                user_id,
                group_id,
                None
            )
        
        else:
            return (
                token,
                None,
                None
            )

                  





import re
from pydantic import (
    BaseModel,
    StrictStr
)
from typing import Literal


class Maintainer(BaseModel):
    layer_type: Literal['maintainer']='maintainer'
    author: StrictStr

    def to_string(self):
        return f'MAINTAINER {self.author}'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('MAINTAINER ', '', line, count=1)
        author = line.strip()

        return Maintainer(
            author=author
        )


import asyncio
import os
import sys

from cocoa.cli import CLI, ImportInstance
from taskex import Env, TaskRunner

from dcrx.image import Image


@CLI.command()
async def compile(
    path: ImportInstance[Image],
    overwrite: bool = False,
):
    runner = TaskRunner(config=Env(MERCURY_SYNC_EXECUTOR_TYPE="thread"))
    loop = asyncio.get_event_loop()

    compilation_images: list[Image] = []

    for image in path.data.values():
        path_exists = await loop.run_in_executor(None, os.path.exists, image.filename)

        if not path_exists or overwrite:
            compilation_images.append(image)
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                f" - Building {image.full_name} at {image.filename}\n",
            )

        else:
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                f" - Failed to compile {image.full_name} to {image.filename} - already exists\n",
            )

    runs = {
        image: runner.run(
            image.to_file,
        )
        for image in compilation_images
    }

    await runner.wait_all([run.token for run in runs.values()])

    for image, run in runs.items():
        if run.error:
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                f" - Failed to compile {image.full_name} to {image.filename} - {run.error}\n",
            )

        else:
            await loop.run_in_executor(
                None,
                sys.stdout.write,
                f" - Created {image.full_name} at {image.filename}\n",
            )

    await runner.shutdown()



import re
from typing import Dict, Literal

from pydantic import BaseModel, StrictBool, StrictStr, constr

SecretMountConfig = Dict[
    Literal[
        "mount_type",
        "id",
        "target",
        "env",
        "required",
        "mode",
        "user_id",
        "group_id",
    ],
    Literal["secret"] | str | bool | Literal[r"^[0-7]*$"],
]


class SecretMount(BaseModel):
    mount_type: Literal["secret"] = "secret"
    id: StrictStr | None = None
    target: StrictStr | None = None
    env: StrictStr | None = None
    required: StrictBool | None = None
    mode: constr(max_length=4, pattern=r"^[0-7]*$") | None = None
    user_id: StrictStr | None = None
    group_id: StrictStr | None = None

    def to_string(self) -> str:
        mount_string = f"--mount=type={self.mount_type}"

        if self.id:
            mount_string = f"{mount_string},id={self.id}"

        if self.target:
            mount_string = f"{mount_string},target={self.target}"

        if self.env:
            mount_string = f"{mount_string},env={self.env}"

        if self.required is not None:
            required = "true" if self.required else False
            mount_string = f"{mount_string},required={required}"

        if self.mode:
            mount_string = f"{mount_string},mode={self.mode}"

        if self.user_id:
            mount_string = f"{mount_string},uid={self.user_id}"

        if self.group_id:
            mount_string = f"{mount_string},gid={self.group_id}"

        return mount_string

    @classmethod
    def parse(cls, line: str):
        options: SecretMountConfig = {"mount_type": "secret"}

        tokens = line.split(",")

        for token in tokens:
            if mount_id := re.search(r"id=(.*)", token):
                options["id"] = re.sub(r"id=", "", mount_id.group(0))

            elif target := re.search(r"target=(.*)", token):
                options["target"] = re.sub(r"target=", "", target.group(0))

            elif env := re.search(r"env=(.*)", token):
                options["env"] = re.sub(r"env=", "", env.group(0))

            elif required := re.search(r"required=(.*)", token):
                required_value = re.sub(r"required=", "", required.group(0))
                options["required"] = True if required_value == "true" else None

            elif mode := re.search(r"mode=[0-7]{4}|[0-7]{3}", token):
                options["mode"] = re.sub(r"mode=", "", mode.group(0))

            elif uid := re.search(r"uid=[0-7]{4}|[0-7]{3}", token):
                options["user_id"] = re.sub(r"uid=", "", uid.group(0))

            elif gid := re.search(r"gid=[0-7]{4}|[0-7]{3}", token):
                options["group_id"] = re.sub(r"gid=", "", gid.group(0))

        return SecretMount(**options)



import re
from typing import Dict, Literal

from pydantic import BaseModel, StrictBool, StrictStr

BindMountConfig = Dict[
    Literal[
        "mount_type",
        "target",
        "source",
        "bind_propataion",
        "from_layer",
        "enable_readonly",
        "enable_readwrite",
    ],
    Literal["bind"]
    | str
    | Literal[
        "rshared",
        "shared",
        "slave",
        "private",
        "rslave",
        "rprivate",
    ]
    | bool,
]


class BindMount(BaseModel):
    mount_type: Literal["bind"] = "bind"
    target: StrictStr
    source: StrictStr | None = None
    bind_propagation: (
        Literal[
            "rshared",
            "shared",
            "slave",
            "private",
            "rslave",
            "rprivate",
        ]
        | None
    ) = None
    from_layer: StrictStr | None = None
    enable_readonly: StrictBool | None = None
    enable_readwrite: StrictBool | None = None

    def to_string(self) -> str:
        mount_string = f"--mount=type={self.mount_type},target={self.target}"

        if self.source:
            mount_string = f"{mount_string},source={self.source}"

        if self.from_layer:
            mount_string = f"{mount_string},from={self.from_layer}"

        if self.enable_readonly:
            mount_string = f"{mount_string},ro"

        if self.enable_readwrite:
            mount_string = f"{mount_string},rw"

        return mount_string

    @classmethod
    def parse(cls, line: str):
        lines = line.split(" ")
        options: BindMountConfig = {"mount_type": "bind"}

        if "readwrite" in lines:
            options["readwrite"] = True
            lines.remove("readwrite")

        elif "rw" in lines:
            options["readwrite"] = True
            lines.remove("rw")

        tokens = line.split(",")

        for token in tokens:
            if target := re.search(r"target=(.*)", token):
                options["target"] = re.sub(r"target=", "", target.group(0))

            elif source := re.search(r"source=(.*)", token):
                options["source"] = re.sub(r"source=", "", source.group(0))

            elif mount_from := re.search(r"from=(.*)", token):
                options["from_layer"] = re.sub(r"from=", "", mount_from.group(0))

            elif readonly := re.search(r"readonly=(.*)", token):
                readonly_value = re.sub(r"readonly=", "", readonly.group(0))
                options["enable_readonly"] = True if readonly_value == "true" else None

            elif re.search(r"readonly", token) or re.search(r"ro", token):
                options["enable_readonly"] = True

            elif readwrite := re.search(r"readwrite=(.*)", token):
                readwrite_value = re.sub(r"readwrite=", "", readwrite.group(0))
                options["enable_readwrite"] = (
                    True if readwrite_value == "true" else None
                )

            elif re.search(r"readwrite", token) or re.search(r"rw", token):
                options["enable_readwrite"] = True

        return BindMount(**options)



import re
from pydantic import BaseModel, StrictStr
from typing import Optional, Literal, Union, Dict
from .mount_types import (
    BindMount,
    CacheMount,
    SecretMount,
    SSHMount,
    TMPFSMount
)


class Run(BaseModel):
    layer_type: Literal["run"]="run"
    command: StrictStr
    mount: Optional[
        BindMount | 
        CacheMount |
        SecretMount |
        SSHMount |
        TMPFSMount
    ]=None
    network: Optional[
        Literal[
            "default", 
            "host", 
            "none"
        ]
    ]=None
    security: Optional[
        Literal[
            "insecure", 
            "sandbox"
        ]
    ]=None

    def to_string(self) -> str:

        run_string = 'RUN'

        if self.mount:
            mount_string = self.mount.to_string()
            run_string = f'{run_string} {mount_string}'

        if self.network:
            run_string = f'{run_string} --network={self.network}'

        if self.security:
            run_string = f'{run_string} --security={self.security}'

        return f'{run_string} {self.command}'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('RUN ', '', line, count=1)
        tokens = line.strip('\n').split(' ')

        options: Dict[str, str | bool | int] = {}
        remainders = []

        for token in tokens:

            if mount := re.search(
                r'--mount=(.*)',
                token
            ):
                
                mount_token = re.sub(
                    r'--mount=',
                    '',
                    mount.group(0)
                )

                options['mount'] = cls._parse_mount(mount_token)
                line = re.sub(
                    mount.group(0),
                    '',
                    line
                )

            elif network := re.search(
                r'--network=(.*)',
                token
            ):
                options['network'] = cls._parse_network(
                    network.group(0)
                )

                line = re.sub(
                    network.group(0),
                    '',
                    line
                )

            elif security := re.search(
                r'--security=(.*)',
                token
            ):
                options['security'] = cls._parse_security(
                    security.group(0)
                )

                line = re.sub(
                    security.group(0),
                    '',
                    line
                )

        return Run(
            command=line,
            **options
        )

    @classmethod
    def _parse_mount(
        cls,
        token: str
    ):
        mount_type = 'bind'

        mount_types: Dict[
            Literal[
                'bind',
                'cache',
                'secret',
                'ssh',
                'tmpfs'
            ],
            BindMount |
            CacheMount |
            SecretMount |
            SSHMount |
            TMPFSMount
        ] = {
            'bind': BindMount,
            'cache': CacheMount,
            'secret': SecretMount,
            'ssh': SSHMount,
            'tmpfs': TMPFSMount
        }


        if mount_match := re.search(
            r'type=(bind|cache|secret|ssh|tmpfs)',
            token
        ):
            mount_type = re.sub(
                r'type=',
                '',
                mount_match.group(0)
            )

            token = re.sub(
                r'type=(bind|cache|secret|ssh|tmpfs)',
                '',
                token
            )

        return mount_types[mount_type].parse(token)
    
    @classmethod
    def _parse_network(
        cls,
        token: str
    ) -> Literal['default', 'host', 'none']:
        if network := re.search(
            r'default|host|none',
            token
        ):
            return network.group(0)
        
        return 'none'
    
    @classmethod
    def _parse_security(
        cls,
        token: str
    ) -> Literal['insecure', 'sandbox']:
        if security := re.search(
            r'insecure|sandbox',
            token
        ):
            return security.group(0)
        
        return 'insecure'

        
        



import io
from typing import Optional, Union


class MemoryFile(io.BytesIO):

    def __init__(
        self, 
        initial_bytes: Union[bytes, bytearray],
        name: Optional[str]=None,
        file_number: int=0
    ) -> None:
        super().__init__(initial_bytes)

        self.name = name
        self.file_number = file_number

    def fileno(self) -> int:
        return self.file_number



import asyncio
import sys

from cocoa.cli import CLI, CLIStyle
from cocoa.ui import (
    Header,
    HeaderConfig,
    Section,
    SectionConfig,
    Terminal,
)

from .compile import compile
from .new import new
from .validate import validate


async def create_header():
    header = Section(
        SectionConfig(height="smallest", width="large", max_height=3),
        components=[
            Header(
                "header",
                HeaderConfig(
                    header_text="dcrx",
                    formatters={
                        "y": [
                            lambda letter, _: "\n".join(
                                [" " + line for line in letter.split("\n")]
                            )
                        ],
                        "l": [
                            lambda letter, _: "\n".join(
                                [
                                    line[:-1] if idx == 2 else line
                                    for idx, line in enumerate(letter.split("\n"))
                                ]
                            )
                        ],
                        "e": [
                            lambda letter, idx: "\n".join(
                                [
                                    line[1:] if idx < 2 else line
                                    for idx, line in enumerate(letter.split("\n"))
                                ]
                            )
                            if idx == 9
                            else letter
                        ],
                    },
                    color="plum_3",
                    attributes=["bold"],
                    terminal_mode="extended",
                ),
            ),
        ],
    )

    terminal = Terminal(
        [
            header,
        ]
    )

    return await terminal.render_once()


@CLI.root(
    compile,
    new,
    validate,
    global_styles=CLIStyle(
        header=create_header,
        flag_description_color="white",
        error_color="pale_violet_red",
        error_attributes=["italic"],
        flag_color="pale_violet_red",
        text_color="plum_3",
        subcommand_color="plum_3",
        indentation=5,
        terminal_mode="extended",
    ),
)
async def dcrx():
    """
    Typesafe Docker image parsing and builds
    """


def run():
    try:
        asyncio.run(CLI.run(args=sys.argv[1:]))

    except (
        KeyboardInterrupt,
        asyncio.CancelledError,
        asyncio.InvalidStateError,
    ):
        pass



from .add import Add
from .arg import Arg
from .cmd import Cmd
from .copy import Copy
from .entrypoint import Entrypoint
from .env import Env
from .expose import Expose
from .healthcheck import Healthcheck
from .label import Label
from .maintainer import Maintainer
from .onbuild import OnBuild
from .run import Run
from .shell import Shell
from .stage import Stage
from .stopsignal import StopSignal
from .user import User
from .volume import Volume
from .workdir import Workdir


import re
from pydantic import (
    BaseModel,
    StrictStr,
)
from typing import Literal


class Label(BaseModel):
    layer_type: Literal["label"]="label"
    name: StrictStr
    value: StrictStr

    def to_string(self) -> str:
        return f'LABEL {self.name}="{self.value}"'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('LABEL ', '', line, count=1)
        name, value = line.strip().split('=')

        return Label(
            name=name,
            value=value
        )


import re
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat
)
from typing import List, Optional, Literal


class Shell(BaseModel):
    layer_type: Literal['shell']='shell'
    executable: StrictStr
    parameters: Optional[
        List[
            StrictBool |
            StrictFloat |
            StrictInt |
            StrictStr
        ]
    ]=None

    def to_string(self):
        shell_command_args = [self.executable]

        if self.parameters and len(self.parameters) > 0:
            shell_command_args.extend([
                parameter if isinstance(
                    parameter,
                    str
                ) else str(parameter) for parameter in self.parameters
            ])

        shell_command = f', '.join(shell_command_args)

        return f'SHELL [ {shell_command} ]'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('SHELL ', '', line, count=1)
        command = [
            arg.strip() for arg in re.sub(
                r'\[|\]', 
                '', 
                line
            ).split(',')
        ]

        parameters: List[str|int|bool|float] | None = None
        if len(command) > 1:
            parameters = command[1:]

        return Shell(
            executable=command[0],
            parameters=parameters
        )


import re
from pydantic import (
    BaseModel,
    StrictInt,
    StrictStr,
    conlist
)

from typing import Literal, List


class Expose(BaseModel):
    layer_type: Literal["expose"]="expose"
    ports: conlist(StrictInt | StrictStr, min_length=1)

    def to_string(self):
        exposed_ports = ' '.join([
            str(port) for port in self.ports
        ])
        return f'EXPOSE {exposed_ports}'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        line = re.sub('EXPOSE ', '', line, count=1).strip()
        ports: List[str] = []

        tokens = line.split(' ')

        for token in tokens:
            if port := re.search(
                r'[0-9]{*}',
                token
            ):
                ports.append(
                    int(port.group(0))
                )

        if len(ports) < 1:
            ports.append(
                line
            )

        return Expose(
            ports=ports
        )
        
        


import re
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat
)

from typing import List, Union, Literal, Dict


class Cmd(BaseModel):
    layer_type: Literal["cmd"]="cmd"
    command: List[Union[StrictStr, StrictInt, StrictFloat, StrictBool]]

    def to_string(self):

        parsed_args = [
            arg.replace('"', '') if isinstance(arg, str) else arg for arg in self.command
        ]

        command = ', '.join([
            f'"{arg}"' for arg in parsed_args
        ])

        return f'CMD [{command}]'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('CMD ', '', line, count=1).strip()
        command = [
            arg.strip() for arg in re.sub(
                r'\[|\]', 
                '', 
                line
            ).split(',')
        ]

        return Cmd(
            command=command
        )


import re
from pydantic import (
    BaseModel,
    StrictInt
)
from typing import Literal


class StopSignal(BaseModel):
    layer_type: Literal['stopsignal']='stopsignal'
    signal: StrictInt

    def to_string(self):
        return f'STOPSIGNAL {self.signal}'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('STOPSIGNAL ', '', line, count=1).strip()
        return StopSignal(
            signal=int(
                re.sub(
                    r'[^0-9]',
                    '',
                    line
                )
            )
        )


import re
from typing import Dict, Literal, Optional

from pydantic import BaseModel, StrictStr


class Stage(BaseModel):
    layer_type: Literal["stage"] = "stage"
    base: StrictStr
    tag: StrictStr | None = None
    platform: Optional[StrictStr] = None
    alias: Optional[StrictStr] = None

    def to_string(self) -> str:
        from_args = ["FROM"]

        if self.platform:
            from_args.append(self.platform)

        if self.tag:
            from_args.append(f"{self.base}:{self.tag}")

        else:
            from_args.append(self.base)

        if self.alias:
            from_args.append(f"AS {self.alias}")

        return " ".join(from_args)

    @classmethod
    def parse(cls, line: str):
        line = re.sub("FROM ", "", line, count=1).strip()
        templated_alias_pattern = re.compile(r" as (.*)| AS (.*)")
        templated_platform_pattern = re.compile(r"--platform=([^\s]+)")

        options: Dict[str, str] = {}

        if platform_match := re.search(templated_platform_pattern, line):
            options["platform"] = re.sub("--platform=", "", platform_match.group(0))

            line = re.sub(templated_platform_pattern, "", line)

        if ":" in line:
            base, tag = line.split(":")

        else:
            base = line
            tag = "latest"

        if alias := re.search(templated_alias_pattern, line):
            options["alias"] = re.sub(r" as | AS ", "", alias.group(0))
            base = re.sub(templated_alias_pattern, "", base)

        if alias := re.search(templated_alias_pattern, tag):
            options["alias"] = re.sub(r" as | AS ", "", alias.group(0))

            tag = re.sub(templated_alias_pattern, "", tag)

        return Stage(base=base, tag=tag, **options)



import re
from typing import Dict, Literal

from pydantic import BaseModel, StrictBool, StrictInt, StrictStr, constr

TMPFSMountConfig = Dict[
    Literal[
        "enable_readonly",
        "enable_readwrite",
        "enable_nosuid",
        "enable_suid",
        "enabled_nodev",
        "enabled_dev",
        "enable_exec",
        "enable_sync",
        "enable_async",
        "enable_dirsync",
        "enable_atime",
        "enable_noatime",
        "enable_diratime",
        "enable_nodiratime",
        "target",
        "mode",
        "user_id",
        "group_id",
        "size",
        "number_inodes",
        "number_blocks",
    ],
    Literal["tmpfs"] | str | bool | Literal[r"^[0-7]*$"] | int,
]


class TMPFSMount(BaseModel):
    mount_type: Literal["tmpfs"] = "tmpfs"
    enable_readonly: StrictBool | None = None
    enable_readwrite: StrictBool | None = None
    enable_nosuid: StrictBool | None = None
    enable_suid: StrictBool | None = None
    enabled_nodev: StrictBool | None = None
    enabled_dev: StrictBool | None = None
    enable_exec: StrictBool | None = None
    enable_sync: StrictBool | None = None
    enable_async: StrictBool | None = None
    enable_dirsync: StrictBool | None = None
    enable_atime: StrictBool | None = None
    enable_noatime: StrictBool | None = None
    enable_diratime: StrictBool | None = None
    enable_nodiratime: StrictBool | None = None
    target: StrictStr
    mode: constr(max_length=4, pattern=r"^[0-7]*$") | None = None
    user_id: StrictStr | None = None
    group_id: StrictStr | None = None
    size: StrictInt | None = None
    number_inodes: StrictInt | None = None
    number_blocks: StrictInt | None = None

    def to_string(self) -> str:
        mount_string = f"--mount=type={self.mount_type},target={self.target}"

        if self.enable_readonly:
            mount_string = f"{mount_string},ro"

        if self.enable_readwrite:
            mount_string = f"{mount_string},rw"

        if self.enable_nosuid:
            mount_string = f"{mount_string},nosuid"

        if self.enable_suid:
            mount_string = f"{mount_string},suid"

        if self.enabled_nodev:
            mount_string = f"{mount_string},nodev"

        if self.enabled_dev:
            mount_string = f"{mount_string},dev"

        if self.enable_exec:
            mount_string = f"{mount_string},exec"

        if self.enable_sync:
            mount_string = f"{mount_string},sync"

        if self.enable_async:
            mount_string = f"{mount_string},async"

        if self.enable_dirsync:
            mount_string = f"{mount_string},dirsync"

        if self.enable_atime:
            mount_string = f"{mount_string},atime"

        if self.enable_noatime:
            mount_string = f"{mount_string},noatime"

        if self.enable_diratime:
            mount_string = f"{mount_string},diratime"

        if self.enable_nodiratime:
            mount_string = f"{mount_string},nodiratime"

        if self.mode:
            mount_string = f"{mount_string},mode={self.mode}"

        if self.user_id:
            mount_string = f"{mount_string},uid={self.user_id}"

        if self.group_id:
            mount_string = f"{mount_string},gid={self.group_id}"

        if self.number_inodes:
            mount_string = f"{mount_string},nr_inodes={self.number_inodes}"

        if self.number_blocks:
            mount_string = f"{mount_string},nr_blocks={self.number_blocks}"

        if self.size:
            mount_string = f"{mount_string},size={self.size}m"

        return mount_string

    @classmethod
    def parse(cls, line: str):
        options: TMPFSMountConfig = {"mount_type": "tmpfs"}

        tokens = line.split(",")

        for token in tokens:
            if target := re.search(r"target=(.*)", token):
                options["target"] = re.sub(r"target=", "", target.group(0))

            elif size := re.search(r"size=[0-9]{*}", token):
                options["size"] = cls._match_size(size.group(0))

            elif readonly := re.search(r"readonly=(.*)", token):
                readonly_value = re.sub(r"readonly=", "", readonly.group(0))
                options["enable_readonly"] = True if readonly_value == "true" else None

            elif re.search(r"readonly", token) or re.search(r"ro", token):
                options["enable_readonly"] = True

            elif readwrite := re.search(r"readwrite=(.*)", token):
                readwrite_value = re.sub(r"readwrite=", "", readwrite.group(0))
                options["enable_readwrite"] = (
                    True if readwrite_value == "true" else None
                )

            elif re.search(r"readwrite", token) or re.search(r"rw", token):
                options["enable_readwrite"] = True

            elif re.search(r"nosuid", token):
                options["enable_nosuid"] = True

            elif re.search(r"suid", token):
                options["enable_suid"] = True

            elif re.search(r"nodev", token):
                options["enable_nodev"] = True

            elif re.search(r"dev", token):
                options["enable_dev"] = True

            elif re.search(r"exec", token):
                options["enable_exec"] = True

            elif re.search(r"async", token):
                options["enable_async"] = True

            elif re.search(r"dirsync", token):
                options["enable_dirsync"] = True

            elif re.search(r"sync", token):
                options["enable_sync"] = True

            if re.search(r"atime", token):
                options["enable_atime"] = True

            elif re.search(r"noatime", token):
                options["enable_noatime"] = True

            elif re.search(r"diratime", token):
                options["enable_diratime"] = True

            elif re.search(r"nodiratime", token):
                options["enable_nodiratime"] = True

            elif mode := re.search(r"mode=[0-7]{4}|[0-7]{3}", token):
                options["mode"] = re.sub(r"mode=", "", mode.group(0))

            elif uid := re.search(r"uid=[0-7]{4}|[0-7]{3}", token):
                options["user_id"] = re.sub(r"uid=", "", uid.group(0))

            elif gid := re.search(r"gid=[0-7]{4}|[0-7]{3}", token):
                options["group_id"] = re.sub(r"gid=", "", gid.group(0))

            elif size := re.search(r"size=\d+", token):
                options["group_id"] = int(re.sub(r"size=", "", size.group(0)))

            elif number_inodes := re.search(r"nr_inodes=\d+", token):
                options["number_inodes"] = int(
                    re.sub(r"nr_inodes=", "", number_inodes.group(0))
                )

            elif number_blocks := re.search(r"nr_blocks=\d+", token):
                options["number_blocks"] = int(
                    re.sub(r"nr_blocks=", "", number_blocks.group(0))
                )

        return TMPFSMount(**options)

    @classmethod
    def _match_size(cls, token: str):
        if size := re.search(r"[0-9]{*}", token):
            return int(size.group(0))



import re
from pydantic import (
    BaseModel,
    StrictStr
)
from typing import Literal

class Workdir(BaseModel):
    layer_type: Literal["workdir"]="workdir"
    path: StrictStr

    def to_string(self) -> str:
        return f'WORKDIR {self.path}'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('WORKDIR ', '', line, count=1).strip()
        return Workdir(
            path=line
        )


import re
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat
)
from typing import Union, Literal, List


class Env(BaseModel):
    layer_type: Literal["env"]="env"
    keys: List[StrictStr]
    values: List[
        StrictStr | StrictInt | StrictBool | StrictFloat
    ]

    def to_string(self) -> str:
        key_value_pairs: List[str] = []
        for key, value in zip(self.keys, self.values):

            if isinstance(value, str):
                value = f'"{value}"'

            key_value_pairs.append(f'{key}={value}')

        key_value_string = '\ \n\t'.join(key_value_pairs)
        return f'ENV {key_value_string}'
        
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('ENV ', '', line, count=1).strip()

        tokens = [
            token.strip(
                '\\'
            ).strip() for token in line.split() if len(token) > 0
        ]

        keys: List[str] = []
        values: List[str] = []

        last_name: str = None

        for token in tokens:

            key, value = None, None
            
            if '=' in token:
                key, value = token.split('=')

            else:

                if len(token.split()) != 2 and last_name is None and len(token) > 0:
                    last_name = token.strip()

                
                elif len(token.split()) != 2 and last_name and len(token) > 0:
                    key = last_name
                    value = token.strip()
                    last_name = None

                else:

                    try:

                        key, value = token.split()

                    except Exception:
                        pass

            if key and value:
                keys.append(
                    key.strip()
                )

                values.append(
                    value.strip()
                )


        return Env(
            keys=keys,
            values=values
        )


import re
from pydantic import (
    BaseModel,
    StrictStr
)

from typing import List, Literal


class Entrypoint(BaseModel):
    layer_type: Literal["entrypoint"]="entrypoint"
    command: List[StrictStr]

    def to_string(self):
        command = ', '.join([
            f'"{arg}"' for arg in self.command
        ])
        
        return f'ENTRYPOINT [{command}]'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        
        line = re.sub('ENTRYPOINT ', '', line, count=1).strip()
        command = [
            arg.strip() for arg in re.sub(
                r'\[|\]', 
                '', 
                line
            ).split(',')
        ]

        return Entrypoint(
            command=command
        )


import re
from enum import Enum
from typing import Literal, Dict, List
from .layers import (
    Add,
    Arg,
    Cmd,
    Copy,
    Entrypoint,
    Env,
    Expose,
    Healthcheck,
    Label,
    Maintainer,
    OnBuild,
    Run,
    Shell,
    Stage,
    StopSignal,
    User,
    Volume,
    Workdir
)


class Directive(Enum):
    ADD='ADD'
    ARG='ARG'
    CMD='CMD'
    COPY='COPY'
    ENTRYPOINT='ENTRYPOINT'
    ENV='ENV'
    EXPOSE='EXPOSE'
    FROM='FROM'
    HEALTHCHECK='HEALTHCHECK'
    LABEL='LABEL'
    MAINTAINER='MAINTAINER'
    ONBUILD='ONBUILD'
    RUN='RUN'
    SHELL='SHELL'
    STOPSIGNAL='STOPSIGNAL'
    USER='USER'
    VOLUME='VOLUME'
    WORKDIR='WORKDIR'


class Directives:

    def __init__(self) -> None:
        self._directives: Dict[
            Literal[
                'ADD',
                'ARG',
                'CMD',
                'COPY',
                'ENTRYPOINT',
                'ENV',
                'EXPOSE',
                'FROM',
                'HEALTHCHECK',
                'LABEL',
                'MAINTAINER',
                'ONBUILD',
                'RUN',
                'SHELL',
                'STOPSIGNAL',
                'USER',
                'VOLUME',
                'WORKDIR'
            ],
            Directive
        ] = {
            'ADD': Directive.ADD,
            'ARG': Directive.ARG,
            'CMD': Directive.CMD,
            'COPY': Directive.COPY,
            'ENTRYPOINT': Directive.ENTRYPOINT,
            'ENV': Directive.ENV,
            'EXPOSE': Directive.EXPOSE,
            'FROM': Directive.FROM,
            'HEALTHCHECK': Directive.HEALTHCHECK,
            'LABEL': Directive.LABEL,
            'MAINTAINER': Directive.MAINTAINER,
            'ONBUILD': Directive.ONBUILD,
            'RUN': Directive.RUN,
            'SHELL': Directive.SHELL,
            'STOPSIGNAL': Directive.STOPSIGNAL,
            'USER': Directive.USER,
            'VOLUME': Directive.VOLUME,
            'WORKDIR': Directive.WORKDIR
        }

        self._layers: Dict[
            Directive,
            Add |
            Arg |
            Cmd |
            Copy |
            Entrypoint |
            Env |
            Expose |
            Healthcheck |
            Label |
            Maintainer |
            OnBuild |
            Run |
            Shell |
            Stage |
            StopSignal |
            User |
            Volume | 
            Workdir
        ] = {
            Directive.ADD: Add,
            Directive.ARG: Arg,
            Directive.CMD: Cmd,
            Directive.COPY: Copy,
            Directive.ENTRYPOINT: Entrypoint,
            Directive.ENV: Env,
            Directive.EXPOSE: Expose,
            Directive.FROM: Stage,
            Directive.HEALTHCHECK: Healthcheck,
            Directive.LABEL: Label,
            Directive.MAINTAINER: Maintainer,
            Directive.ONBUILD: OnBuild,
            Directive.RUN: Run,
            Directive.SHELL: Shell,
            Directive.STOPSIGNAL: StopSignal,
            Directive.USER: User,
            Directive.VOLUME: Volume,
            Directive.WORKDIR: Workdir
        }

        self._directive_names = list(self._directives.keys())
        self._directives_pattern = re.compile(
            '|'.join(self._directive_names)
        )
        self._file_data: List[str] = []
        self._docker_file_lines: List[str] = []

    def __iter__(self):
        for directive_name in self._directives:
            yield directive_name

    def __getitem__(
        self,
        directive_name: Literal[
            'ADD',
            'ARG',
            'CMD',
            'COPY',
            'ENTRYPOINT',
            'ENV',
            'EXPOSE',
            'FROM',
            'HEALTHCHECK',
            'LABEL',
            'MAINTAINER',
            'ONBUILD',
            'RUN',
            'SHELL',
            'STOPSIGNAL',
            'USER',
            'VOLUME',
            'WORKDIR'
        ]
    ):
        return self._directives[directive_name]
    
    def __eq__(
        self,
        directive_name: str
    ) -> bool:
        return directive_name in self._directives
    
    def __contains__(
        self,
        file_line: str
    ):
        return len([
            directive_name for directive_name in self._directives if directive_name in file_line
        ])
    
    def parse(
        self,
        dockerfile: str | List[str] | bytes | List[bytes]
    ) -> List[
        Add |
        Arg |
        Cmd |
        Copy |
        Entrypoint |
        Env |
        Expose |
        Healthcheck |
        Label |
        Maintainer |
        OnBuild |
        Run |
        Shell |
        Stage |
        StopSignal |
        User |
        Volume | 
        Workdir
    ]:
        
        if isinstance(dockerfile, str):
            self._file_data.extend(
                dockerfile.split('\n')
            )

        elif isinstance(dockerfile, bytes):
            self._file_data.extend(
                dockerfile.decode().split('\n')
            )

        else:
            for line in dockerfile:
                if isinstance(line, bytes):
                    line = line.decode()

                self._file_data.append(line)

        self._file_data = [
            line for line in self._file_data if not line.strip().startswith('#') and len(line) > 0
        ]
        
        self._docker_file_lines: List[str] = []

        current_line = ''
        if len(self._file_data) > 0:
            current_line = self._file_data[0]

        for line in self._file_data[1:]:
            if self.from_file_line(line):
                self._docker_file_lines.append(
                    str(current_line)
                )
                current_line = line

            else:
                current_line += line

        self._docker_file_lines.append(current_line)

        docker_directives = []
        for docker_file_line in self._docker_file_lines:
            if directive := self.parse_directive(docker_file_line):
                docker_directives.append(directive)

        return docker_directives
    
    def from_file_line(
        self,
        directive_line: str
    ) -> Literal[
        'ADD',
        'ARG',
        'CMD',
        'COPY',
        'ENTRYPOINT',
        'ENV',
        'EXPOSE',
        'FROM',
        'HEALTHCHECK',
        'LABEL',
        'MAINTAINER',
        'ONBUILD',
        'RUN',
        'SHELL',
        'STOPSIGNAL',
        'USER',
        'VOLUME',
        'WORKDIR'
    ]:

        if matches := re.search(
            self._directives_pattern,
            directive_line
        ):
            return matches.group(0)
        
    def parse_directive(
        self,
        directive_line: str
    ):
        if directive_name := self.from_file_line(directive_line):
            directive_type = self._directives[directive_name]
            return self._layers[directive_type].parse(directive_line)
        



from .bind_mount import BindMount as BindMount
from .bind_mount import BindMountConfig as BindMountConfig
from .cache_mount import CacheMount as CacheMount
from .cache_mount import CacheMountConfig as CacheMountConfig
from .secret_mount import SecretMount as SecretMount
from .secret_mount import SecretMountConfig as SecretMountConfig
from .ssh_mount import SSHMount as SSHMount
from .ssh_mount import SSHMountConfig as SSHMountConfig
from .tmpfs_mount import TMPFSMount as TMPFSMount
from .tmpfs_mount import TMPFSMountConfig as TMPFSMountConfig



import re
from typing import Dict, Literal

from pydantic import BaseModel, StrictBool, StrictStr, constr

CacheMountConfig = Dict[
    Literal[
        "mount_type",
        "id",
        "target",
        "source",
        "from_layer",
        "enable_readonly",
        "enable_readwrite",
        "sharing",
        "mode",
        "user_id",
        "group_id",
    ],
    Literal["cache"]
    | str
    | bool
    | Literal["shared", "private", "locked"]
    | Literal[r"^[0-7]*$"],
]


class CacheMount(BaseModel):
    mount_type: Literal["cache"] = "cache"
    id: StrictStr | None = None
    target: StrictStr
    source: StrictStr | None = None
    from_layer: StrictStr | None = None
    enable_readonly: StrictBool | None = None
    enable_readwrite: StrictBool | None = None
    sharing: Literal["shared", "private", "locked"] = None
    mode: constr(max_length=4, pattern=r"^[0-7]*$") | None = None
    user_id: StrictStr | None = None
    group_id: StrictStr | None = None

    def to_string(self) -> str:
        mount_string = f"--mount=type={self.mount_type},target={self.target}"

        if self.id:
            mount_string = f"{mount_string},id={self.id}"

        if self.source:
            mount_string = f"{mount_string},source={self.source}"

        if self.from_layer:
            mount_string = f"{mount_string},from={self.from_layer}"

        if self.enable_readonly:
            mount_string = f"{mount_string},ro"

        if self.enable_readwrite:
            mount_string = f"{mount_string},rw"

        if self.sharing:
            mount_string = f"{mount_string},sharing={self.sharing}"

        if self.mode:
            mount_string = f"{mount_string},mode={self.mode}"

        if self.user_id:
            mount_string = f"{mount_string},uid={self.user_id}"

        if self.group_id:
            mount_string = f"{mount_string},gid={self.group_id}"

        return mount_string

    @classmethod
    def parse(cls, line: str):
        lines = line.split(" ")
        options: CacheMountConfig = {"mount_type": "cache"}

        if "readonly" in lines:
            options["readonly"] = True
            lines.remove("readonly")

        elif "ro" in lines:
            options["readonly"] = True
            lines.remove("ro")

        tokens = line.split(",")

        for token in tokens:
            if mount_id := re.search(r"id=(.*)", token):
                options["id"] = re.sub(r"id=", "", mount_id.group(0))

            elif target := re.search(r"target=(.*)", token):
                options["target"] = re.sub(r"target=", "", target.group(0))

            elif source := re.search(r"source=(.*)", token):
                options["source"] = re.sub(r"source=", "", source.group(0))

            elif mount_from := re.search(r"from=(.*)", token):
                options["from_layer"] = re.sub(r"from=", "", mount_from.group(0))

            elif readonly := re.search(r"readonly=(.*)", token):
                readonly_value = re.sub(r"readonly=", "", readonly.group(0))
                options["enable_readonly"] = True if readonly_value == "true" else None

            elif re.search(r"readonly", token) or re.search(r"ro", token):
                options["enable_readonly"] = True

            elif readwrite := re.search(r"readwrite=(.*)", token):
                readwrite_value = re.sub(r"readwrite=", "", readwrite.group(0))
                options["enable_readwrite"] = (
                    True if readwrite_value == "true" else None
                )

            elif re.search(r"readwrite", token) or re.search(r"rw", token):
                options["enable_readwrite"] = True

            elif sharing := re.search(r"sharing=(shared|private|locked)", token):
                options["sharing"] = re.search(
                    r"shared|private|locked", sharing.group(0)
                ).group(0)

            elif mode := re.search(r"mode=[0-7]{4}|[0-7]{3}", token):
                options["mode"] = re.sub(r"mode=", "", mode.group(0))

            elif uid := re.search(r"uid=[0-7]{4}|[0-7]{3}", token):
                options["user_id"] = re.sub(r"uid=", "", uid.group(0))

            elif gid := re.search(r"gid=[0-7]{4}|[0-7]{3}", token):
                options["group_id"] = re.sub(r"gid=", "", gid.group(0))

        return CacheMount(**options)



from __future__ import annotations
from dcrx.layers import (
    Add,
    Arg,
    Cmd,
    Copy,
    Entrypoint,
    Env,
    Expose,
    Healthcheck,
    Label,
    Maintainer,
    OnBuild,
    Run,
    Shell,
    Stage,
    StopSignal,
    User,
    Volume,
    Workdir
)
from typing import List, Any, Tuple


class LayerQuery:

    def __init__(
        self,
        layers: List[
            Add |
            Arg |
            Cmd |
            Copy |
            Entrypoint |
            Env |
            Expose |
            Healthcheck |
            Label |
            Maintainer |
            OnBuild |
            Run |
            Shell |
            Stage |
            StopSignal |
            User |
            Volume | 
            Workdir
        ]
    ) -> None:
        self._layers = layers

    def __iter__(self):
        for layer in self._layers:
            yield layer

    def get(
        self,
        attributes: (
            Tuple[str, Any] |
            List[
                Tuple[str, Any]
            ]
        )
    ) -> LayerQuery:
        layers: (
            List[
                Add |
                Arg |
                Cmd |
                Copy |
                Entrypoint |
                Env |
                Expose |
                Healthcheck |
                Label |
                Maintainer |
                OnBuild |
                Run |
                Shell |
                Stage |
                StopSignal |
                User |
                Volume | 
                Workdir
            ]
        ) = []

        if not isinstance(attributes, list):
            attributes = [attributes]

        for layer in self._layers:
            layer_dict = layer.model_dump()

            for attribute, value in attributes:
                layer_attribute = layer_dict.get(attribute)

                if isinstance(layer_attribute, list) and value in layer_attribute:
                    layers.append(layer)
                
                elif (
                    isinstance(layer_attribute, dict) and (
                        value in layer_attribute or value in list(layer_attribute.values()))
                    ):
                    layers.append(layer)

                elif layer_attribute == value:
                    layers.append(layer)

        return LayerQuery(layers)
    
    def results(self):
        return self._layers
    
    def one(self):
        if len(self._layers) > 0:
            return self._layers.pop()
        



import re
from pydantic import (
    BaseModel,
    StrictInt,
    StrictFloat
)
from typing import Literal, Dict, Optional, List
from .cmd import Cmd


class Healthcheck(BaseModel):
    layer_type: Literal["healthcheck"]="healthcheck"
    interval: Optional[StrictInt | StrictFloat]=None
    timeout: Optional[StrictInt| StrictFloat]=None
    start_period: Optional[StrictInt | StrictFloat]=None
    retries: Optional[StrictInt | StrictFloat]=None
    command: Cmd

    def to_string(self) -> str:
        command = self.command.to_string()
        timings: List[str] = []

        if self.interval:
            timings.append(
                f'--interval={self.interval}s'
            )
        
        if self.timeout:
            timings.append(
                f'--timeout={self.timeout}s'
            )

        if self.start_period:
            timings.append(
                f'--start-period={self.start_period}s'
            )

        timings = ' '.join(timings)

        return f'HEALTHCHECK {timings} --retries={self.retries} {command}'
    
    @classmethod
    def parse(
        cls,
        line: str
    ):
        line = re.sub('ADD ', '', line, count=1)

        options: Dict[str, int | Cmd] = {}

        if command := re.search(
            r'CMD(.*)',
            line
        ):
            command_string = command.group(0)
            options['command'] = Cmd.parse(command_string)
            
            line = re.sub(command_string, '', line)

        tokens = line.strip('\n').split(' ')
        
        for token in tokens:
            if interval := re.search(
                r'--interval=[0-9]*s',
                token
            ):
                interval_amount = int(re.sub(
                    r'[^0-9]',
                    '', 
                    interval.group(0)
                ))

                options['interval'] = interval_amount

            elif timeout := re.search(
                r'--timeout=[0-9]*s',
                token
            ):
                timeout_amount = int(re.sub(
                    r'[^0-9]',
                    '', 
                    timeout.group(0)
                ))

                options['timeout'] = timeout_amount
                

            elif start_period := re.search(
                r'--timeout=[0-9]*s',
                token
            ):
                start_period_amount = int(re.sub(
                    r'[^0-9]',
                    '', 
                    start_period.group(0)
                ))

                options['start_period'] = start_period_amount

            elif retries := re.search(
                r'--retries=[0-9]*',
                token
            ):
                
                retries_amount = int(re.sub(
                    r'[^0-9]',
                    '', 
                    retries.group(0)
                ))

                options['retries'] = retries_amount

        return Healthcheck(
            **options
        )
                
                

